from sanic.response import json
from .base import BaseDerivative


class Responses(BaseDerivative):
    @staticmethod
    async def get_access_token_output(request, user, config, instance):
        access_token = await instance.auth.get_access_token(user)

        output = {config.access_token_name: access_token}

        return access_token, output

    @staticmethod
    def get_token_reponse(request,
                          access_token,
                          output,
                          config,
                          refresh_token=None):
        response = json(output)

        if config.cookie_set:
            key = config.cookie_access_token_name
            response.cookies[key] = access_token
            response.cookies[key]['domain'] = \
                config.cookie_domain
            response.cookies[key]['httponly'] = \
                config.cookie_httponly

            if refresh_token and \
                    config.refresh_token_enabled:
                key = config.cookie_refresh_token_name
                response.cookies[key] = refresh_token
                response.cookies[key]['domain'] = \
                    config.cookie_domain
                response.cookies[key]['httponly'] = \
                    config.cookie_httponly

        return response

    @staticmethod
    def extend_authenticate(request,
                            user=None,
                            access_token=None,
                            refresh_token=None):
        return {}

    @staticmethod
    def extend_retrieve_user(request, user=None, payload=None):
        return {}

    @staticmethod
    def extend_verify(request, user=None, payload=None):
        return {}

    @staticmethod
    def extend_refresh(request,
                       user=None,
                       access_token=None,
                       refresh_token=None,
                       purported_token=None,
                       payload=None):
        return {}

    @staticmethod
    def exception_response(request, exception):
        return json({
            'reasons': exception.args[0],
            'exception': exception.__class__.__name__
        }, status=exception.status_code)



from .validators import validate_scopes
from functools import wraps
from inspect import isawaitable
from sanic import Blueprint
from . import exceptions
from sanic.exceptions import add_status_code


def protected(initialized_on=None):
    def decorator(f):
        @wraps(f)
        async def decorated_function(request, *args, **kwargs):
            if initialized_on and isinstance(initialized_on, Blueprint):
                instance = initialized_on
            else:
                instance = request.app

            if request.method == 'OPTIONS':
                response = f(request, *args, **kwargs)
                if isawaitable(response):
                    response = await response
                return response

            try:
                is_authenticated, status, reasons = instance.auth.is_authenticated(
                    request, *args, **kwargs)
            except AttributeError:
                raise add_status_code(500)(exceptions.SanicJWTException(
                    "Authentication instance not found. Perhaps your used "
                    "@protected without passing in a blueprint? "
                    "Try @protected(blueprint)"))
            if is_authenticated:
                response = f(request, *args, **kwargs)
                if isawaitable(response):
                    response = await response
                return response
            else:
                raise add_status_code(status)(exceptions.Unauthorized(reasons))
        return decorated_function
    return decorator


def scoped(scopes,
           require_all=True,
           require_all_actions=True,
           initialized_on=None):
    def decorator(f):
        @wraps(f)
        async def decorated_function(request, *args, **kwargs):
            if initialized_on and isinstance(initialized_on, Blueprint):
                instance = initialized_on
            else:
                instance = request.app

            try:
                is_authenticated, status, reasons = instance.auth.is_authenticated(
                    request, *args, **kwargs)
            except AttributeError:
                raise add_status_code(500)(exceptions.SanicJWTException(
                    "Authentication instance not found. Perhaps your used "
                    "@scoped without passing in a blueprint? "
                    "Try @scoped(..., initialized_on=blueprint)"))
            if is_authenticated:
                # Retrieve the scopes from the payload
                user_scopes = instance.auth.retrieve_scopes(request)
                if user_scopes is None:
                    # If there are no defined scopes in the payload,
                    # deny access
                    is_authorized = False
                    status = 403
                    reasons = "Invalid scope"
                else:
                    is_authorized = await validate_scopes(
                        request, scopes, user_scopes, require_all,
                        require_all_actions, *args, **kwargs)
                    if not is_authorized:
                        status = 403
                        reasons = "Invalid scope"
            else:
                is_authorized = False

            if is_authorized:
                # the user is authorized.
                # run the handler method and return the response
                response = f(request, *args, **kwargs)
                if isawaitable(response):
                    response = await response
                return response
            else:
                raise add_status_code(status)(exceptions.Unauthorized(reasons))
            # return response
        return decorated_function
    return decorator



from . import exceptions
from . import utils
from .base import BaseDerivative
from sanic.views import HTTPMethodView
from sanic.response import json
from sanic.response import text


class BaseEndpoint(BaseDerivative, HTTPMethodView):
    def __init__(self, responses, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.responses = responses


class AuthenticateEndpoint(BaseEndpoint):
    async def options(self, request, *args, **kwargs):
        return text('', status=204)

    async def post(self, request, *args, **kwargs):
        config = self.config
        user = await utils.call(
            self.instance.auth.authenticate, request, *args, **kwargs)

        access_token, output = await self.responses.get_access_token_output(
            request, user, self.config, self.instance)

        if config.refresh_token_enabled:
            refresh_token = await utils.call(
                self.instance.auth.get_refresh_token, request, user)
            output.update({
                config.refresh_token_name: refresh_token
            })
        else:
            refresh_token = None

        output.update(self.responses.extend_authenticate(
            request, user=user, access_token=access_token,
            refresh_token=refresh_token))

        resp = self.responses.get_token_reponse(
            request, access_token, output, refresh_token=refresh_token,
            config=self.config)

        return resp


class RetrieveUserEndpoint(BaseEndpoint):
    async def get(self, request, *args, **kwargs):
        config = self.config
        if not hasattr(self.instance.auth, 'retrieve_user'):
            raise exceptions.MeEndpointNotSetup()

        try:
            payload = self.instance.auth.extract_payload(request)
            user = await utils.call(
                self.instance.auth.retrieve_user, request, payload)
        except exceptions.MissingAuthorizationCookie:
            user = None
            payload = None

        if not user:
            me = None
        else:
            if isinstance(user, dict):
                me = user
            elif hasattr(user, 'to_dict'):
                me = await utils.call(user.to_dict)

        output = {
            'me': me
        }

        output.update(self.responses.extend_retrieve_user(
            request, user=user, payload=payload,))

        resp = json(output)

        if payload is None and config.cookie_set:
            key = config.cookie_access_token_name
            del resp.cookies[key]

        return resp


class VerifyEndpoint(BaseEndpoint):
    async def get(self, request, *args, **kwargs):
        is_valid, status, reason = self.instance.auth.verify(
            request, *args, **kwargs)

        output = {
            'valid': is_valid
        }

        if reason:
            output.update({'reason': reason})

        output.update(self.responses.extend_verify(request,))
        return json(output, status=status)


class RefreshEndpoint(BaseEndpoint):
    async def options(self, request, *args, **kwargs):
        return text('', status=204)

    async def post(self, request, *args, **kwargs):
        # TODO:
        # - Add more exceptions
        payload = self.instance.auth.extract_payload(request, verify=False)
        user = await utils.call(
            self.instance.auth.retrieve_user, request, payload=payload)
        user_id = await self.instance.auth._get_user_id(user)
        refresh_token = await utils.call(
            self.instance.auth.retrieve_refresh_token,
            request=request,
            user_id=user_id)
        if isinstance(refresh_token, bytes):
            refresh_token = refresh_token.decode('utf-8')
        refresh_token = str(refresh_token)
        purported_token = await self.instance.auth \
            .retrieve_refresh_token_from_request(request)

        if refresh_token != purported_token:
            raise exceptions.AuthenticationFailed()

        access_token, output = await self.responses.get_access_token_output(
            request, user, self.config, self.instance)

        output.update(self.responses.extend_refresh(
            request, user=user, access_token=access_token,
            refresh_token=refresh_token, purported_token=purported_token,
            payload=payload,))

        resp = self.responses.get_token_reponse(request, access_token, output,
                                                config=self.config)

        return resp



import logging

from .authentication import Authentication
from .initialization import Initialize
from .initialization import initialize
from .configuration import Configuration
from .responses import Responses
from .decorators import protected
from .decorators import scoped
from .endpoints import BaseEndpoint


logging.getLogger(__name__).addHandler(logging.NullHandler())


__all__ = [
    'Authentication',
    'Initialize',
    'Configuration',
    'Responses',
    'BaseEndpoint',
    'initialize',
    'protected',
    'scoped',
]



import copy
import logging

from . import exceptions
from . import utils


defaults = {
    'access_token_name': 'access_token',
    'algorithm': 'HS256',
    'authorization_header': 'authorization',
    'authorization_header_prefix': 'Bearer',
    'authorization_header_refresh_prefix': 'Refresh',
    'claim_aud': None,
    'claim_iat': False,
    'claim_iss': None,
    'claim_nbf': False,
    'claim_nbf_delta': 0,
    'cookie_domain': '',
    'cookie_httponly': True,
    'cookie_refresh_token_name': 'refresh_token',
    'cookie_set': False,
    'cookie_strict': True,
    'cookie_access_token_name': 'access_token',
    'debug': False,
    'expiration_delta': 60 * 5 * 6,
    'leeway': 60 * 3,
    'refresh_token_enabled': False,
    'refresh_token_name': 'refresh_token',
    'path_to_authenticate': '/',
    'path_to_retrieve_user': '/me',
    'path_to_verify': '/verify',
    'path_to_refresh': '/refresh',
    'private_key': None,
    'scopes_enabled': False,
    'scopes_name': 'scopes',
    'secret': 'This is a big secret. Shhhhh',
    'strict_slashes': False,
    'url_prefix': '/auth',
    'user_id': 'user_id',
    'verify_exp': True,
}

aliases = {
    'cookie_token_name': 'cookie_access_token_name',
    'public_key': 'secret',
}

logger = logging.getLogger(__name__)
config = None


class Configuration:
    def __init__(self, app_config, **kwargs):
        presets = self.extract_presets(app_config)
        self.kwargs = self._merge_aliases(kwargs)
        self.defaults = copy.deepcopy(defaults)
        self.defaults.update(self._merge_aliases(presets))
        self.defaults.update(self.kwargs)

        list(map(self.__map_config, self.defaults.items()))
        self._validate_secret()
        self._validate_keys()
        self._load_keys()

    def __map_config(self, config_item):
        key, value = config_item
        if (not hasattr(self, key) or key in self.kwargs):
            setter_name = 'set_{}'.format(key)
            if hasattr(self, setter_name):
                value = getattr(self, setter_name)()
            setattr(self, key, value)

    def __iter__(self):
        return ((x, getattr(self, x)) for x in self.defaults.keys())  # noqa

    def __repr__(self):
        return str(dict(iter(self)))  # noqa

    def _validate_secret(self):
        logger.debug('validating provided secret')
        if self.secret is None or (
                isinstance(self.secret, str) and self.secret.strip() == ''):
            raise exceptions.InvalidConfiguration(
                'the SANIC_JWT_SECRET parameter cannot be None nor an empty '
                'string')

    def _validate_keys(self):
        logger.debug('validating keys (if needed)')
        if utils.algorithm_is_asymmetric(self.algorithm) and (
            self.private_key is None or (
                isinstance(self.private_key, str) and
                self.private_key.strip() == ''
            )
        ):
            raise exceptions.RequiredKeysNotFound

    def _load_keys(self):
        logger.debug('loading secret and/or keys (if needed)')
        try:
            self.secret = utils.load_file_or_str(self.secret)
            if utils.algorithm_is_asymmetric(self.algorithm):
                self.private_key = utils.load_file_or_str(self.private_key)
        except exceptions.ProvidedPathNotFound as exc:
            if utils.algorithm_is_asymmetric(self.algorithm):
                raise exceptions.RequiredKeysNotFound
            raise exc

    @staticmethod
    def _merge_aliases(config):
        popped = {}
        for k in aliases.keys():
            if k in config:
                popped[aliases[k]] = config.pop(k)
        config.update(popped)
        return config

    @staticmethod
    def extract_presets(app_config):
        """
        Pull the application's configurations for Sanic JWT
        """
        return {
            x.lower()[10:]: app_config.get(x)
            for x in filter(lambda x: x.startswith('SANIC_JWT'), app_config)
        }



from sanic.exceptions import SanicException
from sanic.exceptions import Unauthorized as SanicUnauthorized
from sanic.exceptions import add_status_code


class SanicJWTException(SanicException):
    pass


@add_status_code(401)
class AuthenticationFailed(SanicJWTException):

    def __init__(self, message='Authentication failed.'):
        super().__init__(message)


@add_status_code(400)
class MissingAuthorizationHeader(SanicJWTException):

    def __init__(self, message='Authorization header not present.'):
        super().__init__(message)


@add_status_code(400)
class MissingAuthorizationCookie(SanicJWTException):

    def __init__(self, message='Authorization cookie not present.'):
        super().__init__(message)


@add_status_code(400)
class InvalidAuthorizationHeader(SanicJWTException):

    def __init__(self, message='Authorization header is invalid.'):
        super().__init__(message)


@add_status_code(500)
class AuthenticateNotImplemented(SanicJWTException):

    def __init__(
            self,
            message='Sanic JWT initialized without providing an authenticate '
                    'method.'
    ):
        super().__init__(message)


@add_status_code(500)
class RefreshTokenNotImplemented(SanicJWTException):

    def __init__(self, message="Refresh tokens have not been enabled."):
        super().__init__(message)


@add_status_code(500)
class ScopesNotImplemented(SanicJWTException):

    def __init__(
            self,
            message='Scopes have not been enabled. Initialize with '
                    'add_scopes_to_payload to provide scoping.'
    ):
        super().__init__(message)


@add_status_code(500)
class MissingRegisteredClaim(SanicJWTException):

    def __init__(
            self,
            message='One or more claims have been registered, but your '
                    'SANIC_JWT_HANDLER_PAYLOAD_EXTEND does not supply them. ',
            missing=None):
        if missing:
            message += str(missing)
        super().__init__(message)


@add_status_code(500)
class MeEndpointNotSetup(SanicJWTException):

    def __init__(
            self,
            message='/me endpoint has not been setup. Pass retrieve_user if '
                    'you with to proceeed.'
    ):
        super().__init__(message)


@add_status_code(500)
class InvalidRetrieveUserObject(SanicJWTException):

    def __init__(
            self,
            message='The retrieve_user method should return either a dict or '
                    'an object with a to_dict method.'
    ):
        super().__init__(message)


@add_status_code(500)
class InitializationFailure(SanicJWTException):

    def __init__(
            self,
            message='Sanic JWT was not initialized properly. It must be '
                    'instantiated on a sanic.Sanic or sanic.Blueprint '
                    'instance.'
    ):
        super().__init__(message)


class Unauthorized(SanicJWTException, SanicUnauthorized):

    def __init__(self, message="Auth required."):
        super().__init__(message, scheme="Bearer")


class InvalidClassViewsFormat(SanicJWTException):

    def __init__(
            self,
            message='class_views should follow this format (\'<SOME ROUTE>\', '
                    'ClassInheritedFromHTTPMethodView)'
    ):
        super().__init__(message)


class InvalidConfiguration(SanicJWTException):

    def __init__(self, message=""):
        message = 'An invalid setting was passed to the Sanic JWT ' \
                  'configuration: ' + str(message)
        super().__init__(message)


class InvalidPayload(SanicJWTException):

    def __init__(self, message=""):
        message = 'Payload must be a dictionary with a key mapped to ' \
                  'SANIC_JWT_USER_ID'
        super().__init__(message)


class RequiredKeysNotFound(SanicJWTException):

    def __init__(
            self,
            message='You must provide both (valid) SANIC_JWT_PUBLIC_KEY and '
                    'SANIC_JWT_PRIVATE_KEY when using asymmetric '
                    'cryptographic algorithms like RS*, EC* or PS*'
    ):
        super().__init__(message)


class ProvidedPathNotFound(SanicJWTException):

    def __init__(self,
                 message='The Path object given is not a valid file'):
        super().__init__(message)



from sanic import Blueprint
from sanic import Sanic
from sanic.views import HTTPMethodView
from sanic_jwt import exceptions
from sanic_jwt import endpoints
from sanic_jwt.authentication import Authentication
from sanic_jwt.configuration import Configuration
from sanic_jwt.responses import Responses


def initialize(*args, **kwargs):
    if len(args) > 1:
        kwargs.update({'authenticate': args[1]})
    return Initialize(args[0], **kwargs)


handlers = (
    ('authenticate', (),),
    ('store_refresh_token', ('refresh_token_enabled', ),),
    ('retrieve_refresh_token', ('refresh_token_enabled', ),),
    ('retrieve_user', (),),
    ('add_scopes_to_payload', ('scopes_enabled', ),),
    ('extend_payload', (),),
)

init_classes = (
    'configuration_class',
    'authentication_class',
    'responses_class',
)


class Initialize:
    """Class used to initialize Sanic JWT

    Must be initialized with a keyword argument: `authenticate` that is a
    method that MUST return a user object that when iterated upon returns a
    dictionary, or has a `to_dict` method. The resulting dictionary MUST
    have a key/value for a unique user id.
    """
    authentication_class = Authentication
    configuration_class = Configuration
    responses_class = Responses

    def __init__(self, instance, app=None, **kwargs):
        for class_name in init_classes:
            if class_name in kwargs:
                value = kwargs.pop(class_name)
                setattr(self, class_name, value)

        app = self.__get_app(instance, app=app)
        bp = self.__get_bp(instance)

        self.app = app
        self.bp = bp
        self.kwargs = kwargs
        self.instance = instance

        self.__load_configuration()
        self.__load_responses()
        self.__check_initialization()
        self.__add_class_views()
        self.__add_endpoints()
        self.__initialize_instance()

    def __add_endpoints(self):
        """
        Initialize the Sanic JWT Blueprint and add to the instance initialized
        """
        endpoint_mappings = (
            ('AuthenticateEndpoint', 'authenticate'),
            ('RetrieveUserEndpoint', 'retrieve_user'),
            ('VerifyEndpoint', 'verify'),
            ('RefreshEndpoint', 'refresh'),
        )

        for endpoint in endpoint_mappings:
            self.__add_single_endpoint(*endpoint)

        self.bp.exception(exceptions.SanicJWTException)(
            self.responses.exception_response)

        if not self.instance_is_blueprint:
            url_prefix = self._get_url_prefix()
            self.instance.blueprint(
                self.bp, url_prefix=url_prefix)

    def __add_class_views(self):
        """
        Include any custom class views on the Sanic JWT Blueprint
        """
        config = self.config
        if 'class_views' in self.kwargs:
            class_views = self.kwargs.pop('class_views')

            for route, view in class_views:
                if issubclass(view, HTTPMethodView) and isinstance(route, str):
                    self.bp.add_route(
                        view.as_view(),
                        route,
                        strict_slashes=config.strict_slashes
                    )
                else:
                    raise exceptions.InvalidClassViewsFormat()

    def __check_initialization(self):
        """
        Confirm that required parameters were initialized and report back
        exceptions
        """
        config = self.config
        if hasattr(config, 'refresh_token_enabled') and \
            getattr(config, 'refresh_token_enabled') and (
            not self.kwargs.get('store_refresh_token') or
            not self.kwargs.get('retrieve_refresh_token')
        ):
            raise exceptions.RefreshTokenNotImplemented

        # TODO:
        # - Add additional checks

        # Depracation notices
        if 'SANIC_JWT_HANDLER_PAYLOAD_SCOPES' in self.app.config:
            raise exceptions.InvalidConfiguration(
                "SANIC_JWT_HANDLER_PAYLOAD_SCOPES has been depracated. "
                "Instead, pass your handler method (not an import path) as "
                "initialize(add_scopes_to_payload=my_scope_extender)")

        if 'SANIC_JWT_PAYLOAD_HANDLER' in self.app.config:
            raise exceptions.InvalidConfiguration(
                "SANIC_JWT_PAYLOAD_HANDLER has been depracated. "
                "Instead, you will need to subclass Authentication. ")

        if 'SANIC_JWT_HANDLER_PAYLOAD_EXTEND' in self.app.config:
            raise exceptions.InvalidConfiguration(
                "SANIC_JWT_HANDLER_PAYLOAD_EXTEND has been depracated. "
                "Instead, you will need to subclass Authentication. "
                "Check out the documentation for more information.")

    def __initialize_instance(self):
        """
        Take any predefined methods/handlers and insert them into Sanic JWT
        """
        config = self.config

        # Initialize instance of the Authentication class
        self.instance.auth = self.authentication_class(self.app, config=config)

        if 'authenticate' not in self.kwargs:
            raise exceptions.AuthenticateNotImplemented

        for handler in handlers:
            handler_name, _ = handler
            if handler_name in self.kwargs:
                method = self.kwargs.pop(handler_name)
                setattr(self.instance.auth, handler_name, method)

    def __load_configuration(self):
        """
        Configure settings for the instance in the following order:

        1. Sanic JWT's defaults
        2. Custom Configuration class
        3. Key word arguments passed to Initialize
        """
        config_to_enable = [x for x in handlers if x[1]]
        for config_item in config_to_enable:
            if config_item[0] in self.kwargs:
                list(map(lambda x: self.kwargs.update(
                    {x: True, config_item[0]:
                        self.kwargs.get(config_item[0])}),
                    config_item[1]))

        self.config = self.configuration_class(self.app.config, **self.kwargs)

    def __load_responses(self):
        self.responses = self.responses_class(self.config, self.instance)

    def __add_single_endpoint(self, class_name, path_name):
        view = getattr(endpoints, class_name)
        path_name = getattr(self.config, 'path_to_{}'.format(path_name))
        if self.instance_is_blueprint:
            path_name = self._get_url_prefix() + path_name
            self.instance.add_route(
                view.as_view(config=self.config,
                             instance=self.instance,
                             responses=self.responses), path_name)
        else:
            self.bp.add_route(
                view.as_view(config=self.config,
                             instance=self.instance,
                             responses=self.responses), path_name)

    def _get_url_prefix(self):
        bp_url_prefix = self.bp.url_prefix\
            if self.bp.url_prefix is not None else ''
        config_url_prefix = self.config.url_prefix
        url_prefix = bp_url_prefix + config_url_prefix
        return url_prefix

    @staticmethod
    def __get_app(instance, app=None):
        if isinstance(instance, Sanic):
            return instance
        elif isinstance(instance, Blueprint):
            if app is not None:
                return app
        raise exceptions.InitializationFailure

    @staticmethod
    def __get_bp(instance):
        if isinstance(instance, Sanic):
            return Blueprint('auth_bp')
        elif isinstance(instance, Blueprint):
            return instance
        raise exceptions.InitializationFailure

    @property
    def instance_is_blueprint(self):
        return isinstance(self.instance, Blueprint)



import binascii
import datetime
import inspect
import logging
import os
from pathlib import Path

from . import exceptions

logger = logging.getLogger(__name__)


def generate_token(n=24):
    return str(binascii.hexlify(os.urandom(n)), 'utf-8')


def build_claim_iss(attr, *args, **kwargs):
    return attr


def build_claim_iat(attr, *args, **kwargs):
    return datetime.datetime.utcnow() if attr else None


def build_claim_nbf(attr, config, *args, **kwargs):
    seconds = config.leeway + config.claim_nbf_delta
    return datetime.datetime.utcnow() + datetime.timedelta(
        seconds=seconds
    ) if attr else None


def build_claim_aud(attr, *args, **kwargs):
    return attr


async def call(fn, *args, **kwargs):
    if inspect.iscoroutinefunction(fn) or inspect.isawaitable(fn):
        fn = await fn(*args, **kwargs)
    elif callable(fn):
        fn = fn(*args, **kwargs)
    return fn


def load_file_or_str(path_or_str):
    if isinstance(path_or_str, Path):
        if os.path.isfile(str(path_or_str)):
            logger.debug('reading file \"{}\"'.format(str(path_or_str)))
            return path_or_str.read_text()
        else:
            raise exceptions.ProvidedPathNotFound
    elif isinstance(path_or_str, str):
        if os.path.isfile(path_or_str):
            return Path(path_or_str).read_text()
    return path_or_str


def algorithm_is_asymmetric(algorithm):
    """This is a simple method to verify the need to provide a private key to
    a given ``algorithm``, as `documented by PyJWT
    <https://pyjwt.readthedocs.io/en/latest/algorithms.html>`_

    :param algorithm: the given algorithm, like HS256, ES384, RS512, PS256, etc
    :return: True if algorithm is asymmetric
    """
    return algorithm.lower()[:2] in ('rs', 'es', 'ps')



from sanic_jwt import utils


def validate_single_scope(required, user_scopes, require_all_actions=True):
    def normalize(scope):
        """
        Normalizes and returns tuple consisting of namespace, and action(s)
        """
        parts = scope.split(':')
        return (parts[0], parts[1:])

    required = normalize(required)
    user_scopes = [normalize(x) for x in user_scopes]

    is_valid = False

    for requested in user_scopes:
        if required[0]:
            valid_namespace = required[0] == requested[0]
        else:
            valid_namespace = True

        if required[1]:
            if len(requested[1]) == 0:
                valid_actions = True
            else:
                method = all if require_all_actions else any
                valid_actions = method(x in requested[1] for x in required[1])
        else:
            valid_actions = len(requested[1]) == 0

        is_valid = all([valid_namespace, valid_actions])

        if is_valid:
            break

    return is_valid


async def validate_scopes(
    request,
    scopes,
    user_scopes,
    require_all=True,
    require_all_actions=True,
    *args,
    **kwargs
):
    scopes = await utils.call(scopes, request, *args, **kwargs)

    if not isinstance(scopes, (list, tuple)):
        scopes = [scopes]

    method = all if require_all else any
    return method(
        validate_single_scope(
            x,
            user_scopes,
            require_all_actions=require_all_actions
        ) for x in scopes
    )



import copy
import logging
from datetime import datetime, timedelta

import jwt

from . import exceptions, utils
from .configuration import Configuration

logger = logging.getLogger(__name__)
claim_label = {
    'iss': 'issuer',
    'iat': 'iat',
    'nbf': 'nbf',
    'aud': 'audience',
}


class BaseAuthentication:
    def __init__(self, app, config):
        self.app = app
        self.claims = ['exp']
        if isinstance(config, Configuration):
            self.config = copy.deepcopy(config)
        else:
            msg = 'Sanic JWT was not initialized properly. It did not '\
                'received an instance of Configuration'
            raise exceptions.InitializationFailure(message=msg)

    async def build_payload(self, user, *args, **kwargs):
        if isinstance(user, dict):
            user_id = user.get(self.config.user_id)
        elif hasattr(user, 'to_dict'):
            _to_dict = await utils.call(user.to_dict)
            user_id = _to_dict.get(self.config.user_id)
        else:
            raise exceptions.InvalidRetrieveUserObject()

        return {
            self.config.user_id: user_id,
        }

    async def add_claims(self, payload, *args, **kwargs):
        delta = timedelta(seconds=self.config.expiration_delta)
        exp = datetime.utcnow() + delta
        additional = {
            'exp': exp
        }

        for option in ['iss', 'iat', 'nbf', 'aud', ]:
            setting = 'claim_{}'.format(option.lower())
            attr = getattr(self.config, setting, False)
            if attr:
                method_name = 'build_claim_{}'.format(option)
                method = getattr(utils, method_name)
                additional.update({option: method(attr, self.config)})

        payload.update(additional)

        return payload

    async def extend_payload(self, payload, *args, **kwargs):
        return payload

    async def store_refresh_token(self, *args, **kwargs):
        raise exceptions.RefreshTokenNotImplemented()  # noqa

    async def retrieve_refresh_token(self, *args, **kwargs):
        raise exceptions.RefreshTokenNotImplemented()  # noqa

    async def authenticate(self, *args, **kwargs):
        raise exceptions.AuthenticateNotImplemented()  # noqa

    async def add_scopes_to_payload(self, *args, **kwargs):
        raise exceptions.ScopesNotImplemented()  # noqa


class Authentication(BaseAuthentication):
    def setup_claims(self, *args, **kwargs):
        optional = ['iss', 'iat', 'nbf', 'aud', ]

        for option in optional:
            setting = 'claim_{}'.format(option.upper())
            if getattr(self.config, setting, False):
                self.claims.append(option)

    def _decode(self, token, verify=True):
        secret = self._get_secret()
        algorithm = self._get_algorithm()
        kwargs = {}

        for claim in self.claims:
            if claim != 'exp':
                setting = 'claim_{}'.format(claim.upper())
                value = getattr(self.config, setting, False)
                kwargs.update({claim_label[claim]: value})

        kwargs['leeway'] = int(getattr(self.config, 'leeway', 0))
        if hasattr(self.config, 'claim_aud'):
            kwargs['audience'] = getattr(self.config, 'claim_aud')
        if hasattr(self.config, 'claim_iss'):
            kwargs['issuer'] = getattr(self.config, 'claim_iss')
        return jwt.decode(
            token,
            secret,
            algorithms=[algorithm],
            verify=verify,
            options={
                'verify_exp': self.config.verify_exp
            },
            **kwargs
        )

    def _get_algorithm(self):
        return self.config.algorithm

    async def _get_payload(self, user):
        payload = await utils.call(
            self.build_payload, user)

        if not isinstance(payload, dict) or self.config.user_id not in payload:
            raise exceptions.InvalidPayload

        payload = await utils.call(
            self.add_claims, payload)
        payload = await utils.call(
            self.extend_payload, payload)

        if self.config.scopes_enabled:
            scopes = await utils.call(
                self.add_scopes_to_payload, user)
            if not isinstance(scopes, (tuple, list)):
                scopes = [scopes]
            payload[self.config.scopes_name] = scopes

        missing = [x for x in self.claims if x not in payload]
        if missing:
            raise exceptions.MissingRegisteredClaim(missing=missing)

        return payload

    def _get_secret(self, encode=False):
        # TODO:
        # - Ability to have per user secrets
        if not hasattr(self, '_is_asymmetric'):
            self._is_asymmetric = utils.algorithm_is_asymmetric(
                self.config.algorithm)
        if self._is_asymmetric and encode:
            return self.config.private_key
        return self.config.secret

    def _get_token(self, request, refresh_token=False):
        cookie_token_name_key = 'cookie_access_token_name' \
            if refresh_token is False else \
            'cookie_refresh_token_name'
        cookie_token_name = getattr(self.config, cookie_token_name_key)
        header_prefix_key = 'authorization_header_prefix'
        header_prefix = getattr(self.config, header_prefix_key)

        if self.config.cookie_set:
            token = request.cookies.get(cookie_token_name, None)
            if token:
                return token
            else:
                if self.config.cookie_strict:
                    raise exceptions.MissingAuthorizationCookie()

        header = request.headers.get(
            self.config.authorization_header, None)

        if header:
            try:
                prefix, token = header.split(' ')
                if prefix != header_prefix:
                    raise Exception
            except Exception:
                raise exceptions.InvalidAuthorizationHeader()

            if refresh_token:
                token = request.json.get(
                    self.config.refresh_token_name)

            return token

        raise exceptions.MissingAuthorizationHeader()

    async def _get_refresh_token(self, request):
        return self._get_token(request, refresh_token=True)

    async def _get_user_id(self, user):
        if isinstance(user, dict):
            user_id = user.get(self.config.user_id)
        elif hasattr(user, 'to_dict'):
            _to_dict = await utils.call(user.to_dict)
            user_id = _to_dict.get(self.config.user_id)
        else:
            raise exceptions.InvalidRetrieveUserObject()
        return user_id

    async def get_access_token(self, user):
        payload = await self._get_payload(user)
        secret = self._get_secret(True)
        algorithm = self._get_algorithm()

        return jwt.encode(payload, secret, algorithm=algorithm).decode('utf-8')

    async def get_refresh_token(self, request, user):
        refresh_token = utils.generate_token()
        user_id = await self._get_user_id(user)
        await utils.call(
            self.store_refresh_token,
            user_id=user_id,
            refresh_token=refresh_token,
            request=request)
        return refresh_token

    def is_authenticated(self, request, *args, **kwargs):
        try:
            is_valid, status, reasons = self.verify(request, *args, **kwargs)
        except Exception as e:
            if self.config.debug:
                raise Exception(e)
            else:
                raise exceptions.Unauthorized()

        return is_valid, status, reasons

    def verify(self, request, return_payload=False, verify=True, *args,
               **kwargs):
        token = self._get_token(request)
        is_valid = True
        reason = None

        try:
            payload = self._decode(token, verify=verify)
        except (
            jwt.exceptions.ExpiredSignatureError,
            jwt.exceptions.InvalidIssuerError,
            jwt.exceptions.ImmatureSignatureError,
            jwt.exceptions.InvalidIssuedAtError,
            jwt.exceptions.InvalidAudienceError,
        ) as e:
            is_valid = False
            reason = list(e.args)
            payload = None
            status = 403

        if return_payload:
            return payload

        status = 200 if is_valid else status

        return is_valid, status, reason

    async def retrieve_refresh_token_from_request(self, request):
        return await self._get_refresh_token(request)

    def retrieve_scopes(self, request):
        payload = self.extract_payload(request)
        scopes_attribute = self.config.scopes_name
        return payload.get(scopes_attribute, None)

    def extract_payload(self, request, verify=True, *args, **kwargs):
        payload = self.verify(request, return_payload=True, verify=verify,
                              *args, **kwargs)
        return payload
